# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import argparse
import os.path
import re
import sqlite3
import sys
from   typing import Callable, Iterable, List, Optional, Tuple
import unittest

from .misc import set_indentation

def schapp_main(args:List[str]) -> None:
   """Upgrades a database to target schema given command line parameters"""
   parser= argparse.ArgumentParser(
      prog= os.path.basename(args[0]),
      description="""Schapp applies a target schema to a database, adjusting the
                     database schema to comply with target while migrating existing data.""",
      )
   parser.add_argument('-s', '--schema',
      metavar='<schema script>',
      dest='schema_path',
      required=True,
      help="""Path to SQL file defining the target schema""")
   parser.add_argument('-d', '--db',
      metavar='<database path>',
      dest='db_path',
      required=True,
      help="""Path to database being upgraded""")
   parser.add_argument('upgrade_files', nargs='*', metavar='<upgrade script>',
      help="""Paths to upgrade SQL files that migrate a database with
            schema version 1, schema version 2, etc, respectively, to the target schema""")
   cmdline_args= parser.parse_args(args[1:])

   with sqlite3.connect(cmdline_args.db_path) as conn:
      upgrader= Schapp(conn, cmdline_args.schema_path, cmdline_args.upgrade_files)
      upgrader.perform_upgrade(conn)

class Schapp(object):
   def __init__(self,
        conn: sqlite3.Connection,
        schema_file_path: str,
        upgrade_file_paths: List[str],
        legacy_version_function: Optional[Callable[[sqlite3.Connection], str]]= None):
      """
      Creates a database upgrader. Call `perform_upgrade` method to upgrade the database to the target schema.
      This constructor checks existing versions and finds the target version but does not yet make any changes.

      :param conn: used for inspecting the current version of the database

      :param schema_file_path: points to a script with schapp-complient syntax that defines the target schema.
         The script must contain a comment of the form "-- schapp schema version <version_string>" for schapp to determine
         if any given database needs upgrading. The version_string must start with digits up to the first '.' (or be only
         digits if there is no '.') which as an int form the *major version* which must be at least 1.

      :param upgrade_file_paths: Paths to scripts upgrading from current major version to target schema.
         The first entry upgrades version 1->target, then 2->target etc.
         There must therefore be exactly one element less than the target major version defined inside the schema file.

      :param legacy_version_function: Function that given the open connection returns the database's version.
         Schapp reads the pre-upgrade schema version from the _database_version table in the database. If there is no such
         table, `legacy_version_function` is called (if given) to determine what the current version is. This allows upgrading
         from old versions that were not created with schapp. Without the function a missing _database_version table
         implies that the current database is blank (version "0").
      """
      self._schema_file_path= schema_file_path
      with open(schema_file_path) as stream:
         m= re.search(r"-- schapp schema version ((\d+)([.]\d+)?)", stream.read())
      if m is None:
         raise Exception('Could not find schapp schema version in schema file - expect line of form "-- schapp schema version 0.0"')
      self._target_version= m.group(1)
      self._target_major= int(m.group(2))

      if len(upgrade_file_paths) != self._target_major-1:
         raise Exception("Expected exactly {} upgrade scripts, first one to upgrade schema 1 to schema {}, etc.".format(
                 self._target_major - 1, self._target_major))

      try:
         self._current_version= conn.execute("select current_version from _database_version").fetchall()[0][0]
      except sqlite3.OperationalError as e:
         if legacy_version_function is not None:
            self._current_version= legacy_version_function(conn)
         else:
            self._current_version= "0"
      self._current_major= int(self._current_version.partition(".")[0])
      if self._current_major>self._target_major:
         raise Exception("Current database version is greater than target schema version")
      if 0 < self._current_major < self._target_major:
         self._upgrade_file= upgrade_file_paths[self._current_major - 1]
      else:
         self._upgrade_file= None # upgrade to target schema from same major

   @property
   def requires_upgrade(self) -> bool:
      return self._current_version != self._target_version

   @property
   def preupgrade_major_version(self) -> int:
      return self._current_major

   def perform_upgrade(self, db_conn: sqlite3.Connection) -> None:
      """
      Performs the required upgrade:

      - Creates all new column from schema, leaving all existing ones (even if not in new schema)
      - Drops all constraints
      - Runs given upgrade script which is expected to populate new columns and clear old ones.
      - Drops all old columns, checking they are now empty.
      - Creates all constraints, views, and triggers.

      :param db_conn: this may be a different connection from the one used at construction but must obviously be to
      the same database.
      """
      if not self.requires_upgrade:
         return

      table_defs, trigger_defs, pass_through_defs= read_new_schema(self._schema_file_path)
      version_table= tabledef_from_statementlines([
         "create table _database_version(",
         "  current_version text not null,",
         "  constraint only_one_row check(rowid==1)"
         ");"])
      table_defs.append(version_table)
      existing_tables= set(drop_programmability(db_conn))
      drop_constraints(existing_tables, table_defs, db_conn)

      # Running the appropriate upgrade script to upgrade FROM the current version to head.
      if self._upgrade_file is not None:
         errors= False
         for stmt_lines in read_clean_statements(self._upgrade_file):
            for row in db_conn.execute("\n".join(stmt_lines)):
               print(row)
               errors= True
         if errors: raise Exception("there was integrity checking output")

      create_constraints(table_defs, existing_tables, db_conn)
      create_programmability(trigger_defs, pass_through_defs, db_conn)
      db_conn.execute("pragma foreign_keys= on;")
      violations= db_conn.execute("pragma foreign_key_check;").fetchall()
      if violations:
         raise Exception("FK violations: " + "\n  ".join(repr(row) for row in violations))
      if db_conn.execute("select * from _database_version").fetchall():
         db_conn.execute("update _database_version set current_version= ?", [self._target_version])
      else:
         db_conn.execute("insert into _database_version select ?", [self._target_version])


# [[[cog
# import cornelib
# make_struct= lambda cls,docstr: cornelib.cog_make_struct_type(cog.outl, cls, docstr)
#
# make_struct("Tabledef", '''The should-definition of a table.
#
# Args:
#   name (unicode): table name
#   statement_lines (list[unicode]): The create-statement split into lines.
#   column_types (dict[unicode, unicde]): column_name -> type for all columns in create-statement.
# ''')
#]]]
class Tabledef(object):
   # This is cog-generated code, make changes in the cog-comment above.
   @property
   def name(self):
      return self._name
   @property
   def statement_lines(self):
      return self._statement_lines
   @property
   def column_types(self):
      return self._column_types
   def __repr__(self):
      fields= ", ".join([
            "name="+repr(self._name),
            "statement_lines="+repr(self._statement_lines),
            "column_types="+repr(self._column_types),
         ])
      return "Tabledef({})".format(fields)
   # noinspection PyProtectedMember
   def __eq__(self, other):
      return (isinstance(other, Tabledef)
         and self._name == other._name
         and self._statement_lines == other._statement_lines
         and self._column_types == other._column_types
         )
   def __init__(self,
                name,
                statement_lines,
                column_types,
                ):
      """
      The should-definition of a table.

      Args:
        name (unicode): table name
        statement_lines (list[unicode]): The create-statement split into lines.
        column_types (dict[unicode, unicde]): column_name -> type for all columns in create-statement.
      """
      self._name= name
      self._statement_lines= statement_lines
      self._column_types= column_types
      assert isinstance(name, str)
      assert isinstance(statement_lines, list)
      assert isinstance(column_types, dict)
# [[[end]]]

def start_is_in(s, *beginnings):
   return any(s.lower().startswith(b+" ") for b in beginnings)

def read_clean_statements(script_path):
   """
   Returns iterable of SQL statements from given script file.

   :return: SQL lines forming one statement
   :rtype: str list
   """
   statement= []
   is_block= False
   with open(script_path) as input_stream:
      for line in input_stream:
         if line.strip()=="EOF":
            break
         line= line.rstrip()
         stripped= re.sub(r"\s*(--.*)?$", "", line.strip())
         if stripped or statement:
            statement.append(line)
            if stripped=="begin":
               is_block= True
            if stripped.endswith(";") and (not is_block or stripped=="end;"):
               yield statement
               statement= []
               is_block= False
   if statement:
      raise Exception("incomplete input stream, remaining code: "+" \n".join(statement))

def tabledef_from_statementlines(statement):
   first_line= statement[0]
   if not first_line.endswith("("):
      raise Exception("create table line does not end with '(': "+repr(first_line))
   table_name= first_line[:-1].split()[2]
   column_line_matches= (re.match(r"\s*(\w+)\s+(\w+)\W", line) for line in statement)
   column_types= { m.group(1):m.group(2)
                   for m in column_line_matches
                   if m and m.group(1).lower() not in{"create", "constraint"}
                 }
   return Tabledef(table_name, statement, column_types)

def read_new_schema(schema_script_path):
   """

   Returns: tuple of
     table_defs: Tabledef list
     trigger_defs: statement lines for triggers
     pass_through_defs: statement lines for other objects
   """
   trigger_defs= []
   pass_through_defs= []
   table_defs= []
   for statement in read_clean_statements(schema_script_path):
      first_line= statement[0]
      if start_is_in(first_line, "create table"):
         tabledef= tabledef_from_statementlines(statement)
         table_defs.append(tabledef)
      elif start_is_in(first_line, "create view", "create index", "create unique index"):
         pass_through_defs.append(statement)
      elif start_is_in(first_line, "create triggers"):
         trigger_defs.append(statement)
      else:
         raise Exception('cannot handle "{}"'.format(first_line))
   return table_defs, trigger_defs, pass_through_defs

def drop_programmability(db_conn: sqlite3.Connection) -> Iterable[str]:
   """removes all database objects except tables and returns list of the existing tables
   """
   rs= db_conn.execute("""select type, name
      from sqlite_master
      where not (type='index' and name like 'sqlite^_autoindex^_%' escape '^')""").fetchall()
   for typ, name in rs:
      if typ.lower() in {"trigger", "view", "index"}:
         db_conn.execute("drop {} {};".format(typ, name))
      elif typ.lower()=="table":
         yield name
      else:
         raise Exception("cannot handle {} object".format(typ))

def drop_constraints(existing_tables, new_tabledefs, db_conn: sqlite3.Connection) -> None:
   """
   drops all constraints from existing tables and creates all new columns while leaving all existing columns

   :param existing_tables: names of
   :type existing_tables: str set
   :param new_tabledefs: definitions of all tables to which is being upgrade
   :type new_tabledefs: Tabledef list
   """
   def make_tabledef(name, column_types):
      prev= "create table {}(".format(name)
      for colname,typ in column_types.items():
         yield prev
         prev= "  {} {},".format(colname,typ)
      yield prev.replace(",", ");")
   db_conn.execute("pragma foreign_keys= off;")
   new_column_types= { td.name:td.column_types for td in new_tabledefs }
   for table_name in existing_tables:
      existing_columns_result_set= db_conn.execute("pragma table_info('{}')".format(table_name)).fetchall()
      existing_column_names= [t[1] for t in existing_columns_result_set]
      old_and_new_column_types= {colname:coltype for (_i, colname, coltype, _notnull, _default, _pk) in existing_columns_result_set}
      if table_name in new_column_types:
         old_and_new_column_types.update(new_column_types[table_name])
      statement_lines= make_tabledef(table_name, old_and_new_column_types)
      hide_table(table_name, db_conn)
      recreate_table(table_name, statement_lines, existing_column_names, db_conn)
   for tabledef in new_tabledefs:
      if tabledef.name not in existing_tables:
         statement_lines= make_tabledef(tabledef.name, tabledef.column_types)
         db_conn.execute("\n".join(statement_lines))

def hidden_table_name(original_table_name):
   pfx= "_copy_of_"
   assert not original_table_name.startswith(pfx)
   return pfx+original_table_name

def hide_table(table_name: str, db_conn: sqlite3.Connection) -> None:
   db_conn.execute("alter table {} rename to {};".format(table_name, hidden_table_name(table_name)))

def recreate_table(table_name, statement_lines, copy_columns_list, db_conn: sqlite3.Connection) -> None:
   """
   Creates a table and copies existing data from hidden table.

   :param table_name: name of new table and indirectly of hidden table to be copied
   :type table_name: str
   :param statement_lines: SQL statement that creates the table
   :type statement_lines: str list
   :param copy_columns_list: columns that will be included in the data-copy
   :type copy_columns_list: str iterable
   """
   db_conn.execute("\n".join(statement_lines))
   columns_string= ", ".join(copy_columns_list)
   hidden= hidden_table_name(table_name)
   db_conn.execute("insert into {n}({l})\n  select {l}\n  from {h};".format(n=table_name, l=columns_string, h=hidden))
   db_conn.execute("drop table {};".format(hidden))

def create_constraints(table_defs, existing_tables, db_conn: sqlite3.Connection) -> None:
   new_table_names= set(t.name for t in table_defs)
   for table_name in new_table_names:
      hide_table(table_name, db_conn)
   for table_def in table_defs:
      recreate_table(table_def.name, table_def.statement_lines, table_def.column_types.keys(), db_conn)
   for t in set(existing_tables) - new_table_names:
      db_conn.execute("drop table {};".format(t))

def create_programmability(trigger_defs,
                           pass_through_defs,
                           db_conn: sqlite3.Connection) -> None:
   for statement_lines in pass_through_defs:
      db_conn.execute("\n".join(statement_lines))
   for i, statement_lines in enumerate(trigger_defs):
      _create, _triggers, table_name, columns = statement_lines[0].split(None, 3)
      core_statement= "\n      ".join(statement_lines[1:])
      for insert_update in ("insert", "update of "+columns):
         sql= set_indentation(0, """
            create trigger {table}_{i}{iu}
               before {insert_update} on {table}
               for each row
            {statement}""")
         db_conn.execute(sql.format(i= i,
                                 iu= insert_update[0],
                                 insert_update= insert_update,
                                 table= table_name,
                                 statement= core_statement,
                                 ))
      # getting trigger to check existing data by creating it on dummy table and performing dummmy insert:
      db_conn.execute("create table tmpt(x int);")
      core_statement= re.sub(r"\b(OLD|NEW)\b", table_name, core_statement)
      core_statement= re.sub(r"\bfrom\b", "from {},".format(table_name), core_statement)
      sql= set_indentation(0, """
         create trigger tmptr
            before insert on tmpt
            for each row
         {}""".format(core_statement))
      db_conn.execute(sql)
      db_conn.execute("insert into tmpt values(0);")
      db_conn.execute("drop table tmpt;")


class _Test_Schapp_(unittest.TestCase):
   def test_tabledefs_ending_comma(self):
      script= set_indentation(0, """
         create table T(
            hello int,
            world text not null);""")
      tabledefs,triggers,other = read_new_schema(script)
      self.assertFalse(triggers)
      self.assertFalse(other)
      self.assertEqual(len(tabledefs), 1)
      t= tabledefs[0]
      self.assertEqual(t.name, "T")
      self.assertEqual(t.statement_lines, script.split("\n"))
      self.assertEqual(t.column_types, {'hello':'int', 'world':'text'})



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import itertools
import logging
import os
import shutil
import sqlite3
import time
from   typing import List, Union
import unittest
import unittest.mock

from abus import database
from abus import main

class AbusTestBase(unittest.TestCase):
   def create_file(self, path:str, prime:int, timestamp:float=None) -> float:
      """Creates a file in the home directory

      :param prime: number that somehow determines the file content
      :param path: to create
      :param timestamp: mtime to touch file to, defaults time.time()
      :returns: timestamp of created file
      """
      os.makedirs(os.path.dirname(path), exist_ok=True)
      with open(path,"wb") as f:
         f.write(bytes(n*prime%256 for n in range(8192)))
      if timestamp is None:
         timestamp= time.time() # using the time that the test may have mocked
      os.utime(path, times=(timestamp,timestamp))
      return timestamp

   def homedir_abs(self, *rel_paths) -> Union[str, List[str]]:
      """Returns list of absolute paths from the home directory for the given relative paths.
      If a single path is given, the absolute path is returned directly rather than a list of 1."""
      result= [self.homedir +"/" + rel for rel in rel_paths]
      return result if len(result)!=1 else result[0]

   @property
   def databasepath(self):
      return self.archivedir+"/index.sl3"

   def executesql(self, sql, *params):
      """Runs sql statement on the index database and returns all rows.
      """
      with self.get_direct_db_connection() as conn:
         return conn.execute(sql, params).fetchall()

   def find_files(self, start):
      """Returns direntries for all files in `start`, recursively"""
      q= [start]
      while q:
         entries= list(os.scandir(q.pop())) # prevents iterator leakage
         yield from (e for e in entries if e.is_file())
         q.extend(e.path for e in entries if e.is_dir())

   def get_direct_db_connection(self):
      conn= sqlite3.connect(self.databasepath)
      conn.isolation_level= None # autocommit
      conn.execute("PRAGMA synchronous=OFF")
      return conn

   def remove_dir_contents(self, path):
      for direntry in os.scandir(path):
         if direntry.is_dir():
            shutil.rmtree(direntry.path)
         elif direntry.is_file():
            os.unlink(direntry.path)

   def setup_directories(self):
      """
      Sets up empty home, archive, and restore dirs, config file and empty database.

      :rtype: None
      """
      logging.basicConfig(level=logging.DEBUG)
      candidates_for_tmp= ["C:/tmp", "C:/temp"]
      if "TEMP" in os.environ: candidates_for_tmp.append(os.environ["TEMP"])
      for tempdir in candidates_for_tmp:
         if os.path.isdir(tempdir):
            break
      else:
         raise Exception("Could not find a temporary directory for tests")

      self.password= "Sesam, öffne dich!"
      self.homedir= tempdir + "/abushome"
      self.otherdir= tempdir + "/abusother"
      self.archivedir= tempdir + "/abusarchive"
      self.restoredir= tempdir + "/abusrestore"
      os.makedirs(self.homedir, exist_ok=True)
      os.makedirs(self.archivedir, exist_ok=True)
      os.makedirs(self.restoredir, exist_ok=True)
      os.makedirs(self.otherdir, exist_ok=True)
      os.chdir(self.restoredir)

      # empty them
      self.remove_dir_contents(self.homedir)
      self.remove_dir_contents(self.archivedir)
      self.remove_dir_contents(self.restoredir)
      self.remove_dir_contents(self.otherdir)

      self.configfile= self.otherdir + "/abusconfig.cfg"
      self.write_config_file()

      with database.connect(self.databasepath, self.archivedir, allow_create=True):
         pass

   def write_config_file(self, *extra_param_lines):
      with open(self.configfile, "w", encoding="utf-8") as f:
         print("logfile", self.otherdir + "/abuslog.txt", file=f)
         print("archive", self.archivedir, file=f)
         print("password", self.password, file=f)
         print("\n".join(extra_param_lines), file=f)
         print("[include]", file=f)
         print(self.homedir, file=f)

   def run_backup(self, expected_return_code:int=0) -> None:
      """
      Calls main() for backup with the config file and checks the return code.

      :param expected_return_code: asserts this is returned by main()
      """
      self.run_main("--backup", expected_return_code)

   def run_main(self, action_option: str, expected_return_code:int=0) -> None:
      """
      Calls main() with given action_option and the config file and checks the return code.

      :param action_option: E.g. --restore
      :param expected_return_code: asserts this is returned by main()
      """
      rc= main.main(["test", "-f", self.configfile, action_option])
      self.assertEqual(rc, expected_return_code)

   def setup_backup_with_well_known_checksums(self):
      self.setup_directories()
      self.expected_backups= set() # contains (path rel to homedir, backup filename) pairs

      path= self.homedir_abs("my_valueable_file")
      with open(path, "w", encoding="utf-8") as f:
         for n in range(400):
            for p in range(10):
               print(n**p, file=f, end="")
            print(file=f)
      self.expected_backups.add((path,"73672933c00ab3cd730c8715a392ee6dee9ba2c0a8e5e5d07170b6544b0113ef.z"))

      path= self.homedir_abs("my_little_file")
      with open(path, "wb") as f:
         f.write(bytearray(range(256)))
      self.expected_backups.add((path,"40aff2e9d2d8922e47afd4648e6967497158785fbd1da870e7110266bf944880"))

      path= self.homedir_abs("subdir_a/I_am_in_a.tif")
      self.create_file(path, 31)
      self.expected_backups.add((path,"498630374d56ea4c53c5baaee00b630be09ca5dbe678b11b1a81b37f1058635f"))

      path= self.homedir_abs("subdir_b/I_am_in_b.bin")
      self.create_file(path, 97)
      self.expected_backups.add((path,"1244d0711c6534ee79a6d3b82cea33b76e766e46cbca75791ac9fa3a30e365f3.z"))

      rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 0)

   def setup_multiple_backups(self):
      self.setup_directories()
      def primes():
         prev= [2]
         for c in itertools.count(3,2):
            for p in prev:
               if p*p>c:
                  yield c
                  prev.append(c)
                  break
               if c%p==0: break
      primes= primes().__iter__()

      all_paths= self.homedir_abs("a", "b", "s/a", "s/b", "t/a", "t/b")
      n= len(all_paths)*3//4
      for i in range(-3,0):
         subtract_time= -i*86400
         change_paths= all_paths if i==0 else all_paths[:n] if i==1 else all_paths[-n:]
         for path in change_paths:
            self.create_file(path, next(primes), time.time()-subtract_time-1200)
         with self.turn_back_the_time(subtract_time+600):
            rc= main.main(["test", "-f", self.configfile, "--backup"])
         self.assertEqual(rc, 0)

   def turn_back_the_time(self, seconds:float):
      """Returns a context manager for a mocked time.time() which is `seconds` behind the real time."""
      orig_time= time.time
      mock= lambda: orig_time() - seconds
      return unittest.mock.patch('time.time', side_effect=mock)



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import gzip
import os
import shutil
import sqlite3
import time
import unittest.mock
from abus import backup
from abus import config
from abus import database
from abus import crypto
from abus import main
from abus import purge
from abus.testbase import AbusTestBase

class BackedUpHomedirTests(AbusTestBase):
   def test_simple_end_to_end_backup(self):
      self.setup_backup_with_well_known_checksums()

      # checking archive contents matches DB
      subdir,run_name = self.executesql("select archive_dir, run_name from run")[0]
      index_file_path_rel= subdir+"/"+run_name+".lst"
      expected_backupfiles= {"index.sl3", index_file_path_rel, run_name+".gz"}
      entries= self.executesql("""
         select path, archive_dir, location.checksum, is_compressed
         from content
            join location on location.checksum = content.checksum""")
      expected_backupfiles.update(a+"/"+c+(".z" if z else "") for p,a,c,z in entries)
      expected_backupfiles_abs= {self.archivedir+"/"+rel for rel in expected_backupfiles}
      actual= set(direntry.path.replace("\\","/")
                  for direntry in self.find_files(self.archivedir))
      self.assertEqual(actual, expected_backupfiles_abs)

      # checking archive contents matches what has been backed up
      # rs= e.g. ('C:/tmp/abushome/my_little_file', '40aff2e9d2d8922e47afd4648e6967497158785fbd1da870e7110266bf944880', 0)
      from_db= {(p, c+".z" if z else c) for p,a,c,z in entries}
      from_test= set(self.expected_backups)
      self.assertEqual(from_db, from_test)

      # checking that index files matches
      with crypto.open_txt(self.archivedir+"/"+index_file_path_rel, "r", self.password) as f:
         index_entries= set(l.strip() for l in f)
      dbentries= set("{} {} {}".format(*row)
                     for row in self.executesql("select checksum, timestamp, path from content"))
      self.assertEqual(index_entries, dbentries)

      # checking that contents file matches
      with gzip.open(self.archivedir+"/"+run_name+".gz", "rt", encoding="UTF-8") as stream:
         lines= {l.strip() for l in stream}
      self.assertEqual(lines, expected_backupfiles)

      # decrypting backup
      my_litte_backup= [e for e in expected_backupfiles_abs if "/40aff2e" in e]
      self.assertEqual(len(my_litte_backup), 1)
      self.assertEqual(os.stat(my_litte_backup[0]).st_size, 256+80) # 80 for salt, init vector, and checksum
      with crypto.open_bin(my_litte_backup[0], "r", self.password) as f:
         blk= f.read(8192)
         self.assertEqual(len(blk), 256)
         self.assertEqual(blk, bytearray(range(256)))

   def test_error_is_written_to_runfile(self):
      self.setup_directories()
      path= self.homedir_abs("some_file.txt")
      self.create_file(path, 23)
      with unittest.mock.patch('abus.backup.make_backup_copy', side_effect=RuntimeError):
         self.run_backup(expected_return_code=1)
      rs= self.executesql("select run_name, archive_dir from run")
      self.assertEqual(len(rs), 1)
      run_name, archive_dir = rs[0]
      index_file_path= self.archivedir+"/"+rs[0][1]+"/"+rs[0][0]+".lst"
      with crypto.open_txt(index_file_path, "r", self.password) as stream:
         contents= set(stream)
      error_line= "error 0 {}\n".format(path)
      self.assertIn(error_line, contents)
      self.assertTrue(os.path.isfile(self.archivedir+"/"+run_name+".gz"))

   def test_deletions(self):
      self.setup_directories()
      def check_deletions(*expected):
         actual= set(path for (path,) in self.executesql("select path from deletion"))
         self.assertEqual(actual, set(expected))

      file_a, file_b, error_dir = self.homedir_abs("file_a", "file_b", "subdir")

      with self.turn_back_the_time(4*86400):
         self.create_file(file_a, 17)
         self.create_file(file_b, 27)
         self.run_backup()
      check_deletions()

      with self.turn_back_the_time(3*86400):
         os.unlink(file_a)
         self.run_backup()
      check_deletions(file_a)

      with self.turn_back_the_time(2*86400):
         self.create_file(file_a, 97)
         os.unlink(file_b)
         self.run_backup()
      check_deletions(file_b)

      # when there is an error that risks not seeing some files
      # then files most not be marked deleted even if they are
      # but files that were seen must still be undeleted
      orig_scandir= os.scandir
      def mock_scandir(path):
         if path==error_dir:
            raise OSError("don't feel like reading that")
         return orig_scandir(path)
      with self.turn_back_the_time(86400):
         os.mkdir(error_dir)
         os.unlink(file_a)
         self.create_file(file_b, 73)
         with unittest.mock.patch("os.scandir", side_effect=mock_scandir):
            self.run_backup(expected_return_code=1)
      check_deletions()

      # when there is no error any more
      # then missing file must be marked as deleted
      self.run_backup()
      check_deletions(file_a)

   def test_changing_file_gets_backed_up_after_retry(self):
      self.setup_directories()
      path= self.homedir_abs("some_file.txt")
      self.create_file(path, 23)
      def make_backup_copy_sideeffect(path, expected_checksum, backup_path, open_dst_function, password):
         self.create_file(path, 97) # 1244d0711c6534ee79a6d3b82cea33b76e766e46cbca75791ac9fa3a30e365f3
      with unittest.mock.patch('time.sleep'):
         with unittest.mock.patch('abus.backup.make_backup_copy',
                                  wraps=backup.make_backup_copy,
                                  side_effect=make_backup_copy_sideeffect):
            rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 0)
      rs= self.executesql("select run_name, archive_dir from run")
      self.assertEqual(len(rs), 1)
      _run_name, _archive_dir = rs[0]
      rs= self.executesql("""select location.checksum
         from content join location on location.checksum = content.checksum""")
      self.assertEqual(len(rs), 1)
      self.assertEqual(rs[0][0], "1244d0711c6534ee79a6d3b82cea33b76e766e46cbca75791ac9fa3a30e365f3")

   def test_always_changing_file_causes_error_after_5_attempts(self):
      self.setup_directories()
      primes= iter([17,19,23,29,31,37,41])
      path= self.homedir_abs("some_file.txt")
      self.create_file(path, next(primes))
      def make_backup_copy_sideeffect(path, expected_checksum, backup_path, open_dst_function, password):
         self.create_file(path, next(primes))
      with unittest.mock.patch('time.sleep'):
         with unittest.mock.patch('abus.backup.make_backup_copy',
                                  wraps=backup.make_backup_copy,
                                  side_effect=make_backup_copy_sideeffect) as mock:
            rc= main.main(["test", "-f", self.configfile, "--backup"])
            self.assertEqual(mock.call_count, 5)
      self.assertEqual(rc, 1)
      rs= self.executesql("select * from location")
      self.assertEqual(rs, [])

   def test_bad_include_entry_does_not_stop_other_backups(self):
      """bug: read error on c:/users/*/documents causes unhandled exception and aborts whole backup"""
      self.setup_directories()
      # Appending bad [include]-entry happens to run it first and used to prevent homedir from being backed up
      with open(self.configfile, "a") as stream:
         stream.write("C:/Users/*/Desktop\n")
      path= self.homedir_abs("some_file.txt")
      self.create_file(path, 23)
      rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 1)
      with self.get_direct_db_connection() as conn:
         rows= [conn.execute("select * from "+t).fetchall() for t in("content", "run", "location", "run")]
      self.assertEqual(len(rows[0]), 1)
      self.assertEqual(len(rows[1]), 1)
      self.assertEqual(len(rows[2]), 1)
      self.assertEqual(len(rows[3]), 1)

   def test_archive_dirs_to_use(self):
      def go(initial_usage: dict, expected_result: list):
         i= iter(backup.ArchiveDirsToUse._sequence(initial_usage))
         for archive_dir, n, skip in expected_result:
            actual= [next(i) for _ in range(n)]
            expect= [archive_dir]*n
            self.assertEqual(actual, expect)
            for _ in range(skip):
               next(i)
      go({},
         [("00", 100, 0), ("01", 100, 9800), ("00/01", 100, 0), ("01/01", 100, 9800), ("00/02", 100, 979800),
          ("99/99", 100, 0),
          ("00/00/01", 100, 0)])
      go({"00": 100, "45": 30},
         [("45", 70, 0), ("01", 100, 9700), ("00/01", 100, 0), ("01/01", 100, 9800), ("00/02", 100, 0)])

   def test_indexdb_in_different_location(self):
      self.setup_directories()
      path= self.homedir_abs("some_file.txt")
      self.create_file(path, 79)
      db_path= self.otherdir+"/index_database"
      shutil.move(self.databasepath, db_path)
      self.write_config_file("indexdb " +db_path)
      rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 0)
      with sqlite3.connect(db_path) as conn:
         content= conn.execute("select * from content").fetchall()
         subdir,run_name = conn.execute("select archive_dir, run_name from run").fetchall()[0]
      self.assertEqual(len(content), 1)
      self.assertFalse(os.path.exists(self.databasepath))

      content_file_path= self.archivedir+"/"+run_name+".gz"
      with gzip.open(content_file_path, "rt", encoding="UTF-8") as stream:
         lines= {l.strip() for l in stream}
      lines.remove(run_name+".gz")
      lines.remove(subdir+"/"+run_name+".lst")
      self.assertEqual(len(lines), 1)
      self.assertRegex(lines.pop(), "[a-z0-9]{64}")

      # bug: Exception in os.path.relpath if db on a different drive.
      os.unlink(content_file_path)
      cfg= config.Configuration(["test", "-f", self.configfile, "--backup"])
      with database.connect(cfg.database_path, cfg.archive_root_path, allow_create=False) as db:
         cfg.database_path= "X:/indexdb" # database_path no linger used, we can lie about it now
         with self.assertRaises(ValueError):
            os.path.relpath(cfg.database_path, cfg.archive_root_path)
         purge.write_content_file(cfg, db)

   def test_db_u64_ino_dev(self):
      self.setup_directories()
      path, st_dev, st_ino, mtime, ctime = "C:/tmp/some_file.txt", 2**63, 2**63, 1500000000.0, 0.0
      with database.connect(self.databasepath, self.archivedir) as db:
         db.remember_file_metadata(path, st_dev, st_ino, mtime, ctime)
         self.assertTrue(db.is_file_unchanged(path, st_dev, st_ino, mtime, ctime))

   def test_no_repeating_entries(self):
      self.setup_directories()
      day= 86400
      t0= round(time.time() - 6*day-1000)
      t1= t0 + 2*day

      file_a, file_b = self.homedir_abs("file_a.txt", "file_b.txt")
      self.create_file(file_a, 31, t0)
      with self.turn_back_the_time(6*day):
         rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 0)

      self.create_file(file_a, 97, t1)
      with self.turn_back_the_time(4*day):
         rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 0)

      self.create_file(file_a, 31, t0) # this should appear in the content
      with self.turn_back_the_time(2*day):
         rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 0)

      os.rename(file_a, file_b)
      self.create_file(file_a, 31, t0) # this should not appear despite changed ino
      rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 0)

      entries= self.executesql("""select path, timestamp, checksum
         from content
         order by path, run_name""")
      expect= [
         (file_a, t0, "498630374d56ea4c53c5baaee00b630be09ca5dbe678b11b1a81b37f1058635f"),
         (file_a, t1, "1244d0711c6534ee79a6d3b82cea33b76e766e46cbca75791ac9fa3a30e365f3"),
         (file_a, t0, "498630374d56ea4c53c5baaee00b630be09ca5dbe678b11b1a81b37f1058635f"),
         (file_b, t0, "498630374d56ea4c53c5baaee00b630be09ca5dbe678b11b1a81b37f1058635f"),
         ]
      self.assertEqual(entries, expect)



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import os
import time
from   typing import Tuple
from abus import database
from abus import main
from abus.testbase import AbusTestBase

DAY= 86400

class PurgeTests(AbusTestBase):
   def test_purge_e2e(self):
      self.setup_directories()
      def checkdb(expected):
         data= self.executesql("""
            select run_name, path, timestamp, content.checksum, location.is_compressed
            from content
               left join location on location.checksum = content.checksum""")
         data= {(r+".lst",p,t,c,z) for r,p,t,c,z in data}
         self.assertEqual(data, expected)
      def get_new_archive_files(*existing_filenames) -> Tuple[set,set,set]:
         """Returns all new backup files give the one expected to exist already.

         :return: (new run filenames, new compressed backup filenames, new uncompressed)
         """
         all_files= set(direntry.name for direntry in self.find_files(self.archivedir))
         content_file= {n for n in all_files if n.endswith(".gz")}
         relevant= all_files -set(existing_filenames) -{"index.sl3", elt(content_file)}
         runs= set(n for n in relevant if n.endswith(".lst"))
         compressed= set(n for n in relevant if n.endswith(".z"))
         uncompressed= relevant -runs -compressed
         return runs, compressed, uncompressed
      def elt(s):
         """returns the only element from set"""
         self.assertEqual(len(s), 1)
         for e in s: return e

      day= 86400
      path_a,path_b = self.homedir_abs("path_a.txt", "path_b.gz")
      # -11 -10    -9 -8   -7  -2  0    t relative to now in days
      #  r0      r2-7 r1 r3-7  r2 r3    runs
      #  a0  a1                         both a versions come into r2's >7days slot, a0 is purged in r2
      #  b0           b1                b1 is in different r2 slot, but is purged in r3
      with self.turn_back_the_time(11*day):
         t0= time.time()
         self.create_file(path_a, 23, t0)
         self.create_file(path_b, 11, t0)
         rc= main.main(["test", "-f", self.configfile, "--backup"])
         self.assertEqual(rc, 0)
      runs, compressed, uncompressed = get_new_archive_files() # -> file_a0, file_b0, run0
      file_a0= elt(compressed)
      file_b0= elt(uncompressed)
      run0= elt(runs)
      checkdb({(run0, path_a, t0, file_a0[:64], 1),
               (run0, path_b, t0, file_b0, 0),
               })

      with self.turn_back_the_time(8*day):
         ta1= time.time()-2*day
         tb1= time.time()
         self.create_file(path_a, 17, ta1)
         self.create_file(path_b, 59, tb1)
         rc= main.main(["test", "-f", self.configfile, "--backup"])
         self.assertEqual(rc, 0)
      runs, compressed, uncompressed = get_new_archive_files(run0, file_a0, file_b0) # -> run1, file_a1, file_b1
      file_a1= elt(compressed)
      file_b1= elt(uncompressed)
      run1= elt(runs)
      checkdb({(run0, path_a, t0, file_a0[:64], 1),
               (run0, path_b, t0, file_b0, 0),
               (run1, path_a, ta1, file_a1[:64], 1),
               (run1, path_b, tb1, file_b1, 0),
               })

      with self.turn_back_the_time(2*day):
         rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 0)
      # file_a0 purged
      runs, compressed, uncompressed = get_new_archive_files(run0, file_b0, run1, file_a1, file_b1) # -> run2
      self.assertEqual(compressed, set())
      self.assertEqual(uncompressed, set())
      _run2= elt(runs)
      checkdb({(run0, path_b, t0, file_b0, 0),
               (run1, path_a, ta1, file_a1[:64], 1),
               (run1, path_b, tb1, file_b1, 0),
               })

      rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 0)
      # file_b0, run0, run2 purged
      runs, compressed, uncompressed = get_new_archive_files(run1, file_a1, file_b1) # -> run3
      self.assertEqual(compressed, set())
      self.assertEqual(uncompressed, set())
      _run3= elt(runs)
      checkdb({(run1, path_a, ta1, file_a1[:64], 1),
               (run1, path_b, tb1, file_b1, 0),
               })

   def test_db_get_purgeable_backups(self):
      self.setup_directories()
      rounders= [(1, 3), (4, 9)]
      aged_based_data= [
         # rounders:    0   0   0    0    4    4    4  4  4    4    4    4    1   1    1    1    1    1    1 <-- matches era calculation below
         ("path_a", (1000,20  ,  10, 9.1, 8.9, 8.1, 7.9,   6,   5, 4.1, 3.9, 3.1, 2.9,1.9, 1.5, 1.1, 0.9, 0.4, 0.2)),
         ("file_b", (None,19.5)),
         ("file_c", (1001,30  ,None,None,None,None,None,None,None,None,None,None,None,2  )),
         ("file_1d",(1002,31)), # shared checksums
         ("file_2d",(1002,32)),
         ("file_1e",(1003,33)), # shared checksums
         ("file_2e",(1003,None,None,None,None,None,   8)),
      ]
      now= time.time()
      timestamp_lists= {path:[None if age is None else now-age*86400 for age in ages] for path,ages in aged_based_data}
      checksum_lists= {path:[None if t is None else path[-1] +("0"*64 +str(t))[-63:] for t in timestamps]
                       for path,timestamps in timestamp_lists.items()}
      # maps checksums to 3 periods defined by rounders. Period id is maximum number of slots in that period (depending on time of day)
      eras= {}
      for path in timestamp_lists:
         for i, checksum in enumerate(checksum_lists[path]):
            if not checksum: continue
            eras[checksum]= 1 if i<4 else 3 if i<12 else 4
      with database.connect(self.databasepath, self.archivedir) as db:
         runs= [db._make_runname(t+65) for t in timestamp_lists["path_a"]]
         all_checksums= set()
         for path in timestamp_lists:
            for run_name, timestamp, checksum in zip(runs, timestamp_lists[path], checksum_lists[path]):
               if not timestamp: continue
               db.add_backup_entry(run_name, path, timestamp, checksum)
               all_checksums.add(checksum)
         for checksum in all_checksums:
               db.remember_archivedir(checksum, checksum[-2:], True)
         del all_checksums

         result= db.get_purgeable_backups(rounders)
         self.assertTrue(all(a==c[-2:] for c,a,p,t in result)) # correct checksum

         result_checksums= set(c for c,a,p,t in result)

         # the latest versions must never be purged
         latest= set(cl[-1] for cl in checksum_lists.values())
         self.assertEqual(latest & result_checksums, set())

         # file_a
         keep= set(checksum_lists["path_a"]) - result_checksums
         # grouping by era, there must be at most as many versions to keep as slots in era.
         # cannot check "one per slot" directly because slot boundaries depend on time of day
         keep_by_era= {era: set() for era in eras.values()}
         for checksum in keep:
            keep_by_era[eras[checksum]].add(checksum)
         for era in keep_by_era:
            with self.subTest(contents=keep_by_era[era]):
               self.assertLessEqual(len(keep_by_era[era]), era)

         # file_c - one backup in oldest slot is always kept
         purged= [c for c in result_checksums if c.startswith("c")]
         self.assertEqual(purged, [checksum_lists["file_c"][0]]) # 30 and 2 stay

         # file_?d - oldest gets deleted because both paths don't need it any more
         purged= [c for c in result_checksums if c.startswith("d")]
         self.assertEqual(purged, [checksum_lists["file_1d"][0]])
         self.assertEqual(purged, [checksum_lists["file_2d"][0]])

         # file_?e - no deletion because the other still needs it
         purged= [c for c in result_checksums if c.startswith("e")]
         self.assertEqual(purged, [])

   def test_current_run_does_not_get_purged(self):
      """With the change that files are not added to content if the they have not changed since
      the last run, that current run may actually be empty, but must not be deleted if it is."""
      self.setup_directories()
      path= self.homedir_abs("some_file.txt")
      self.create_file(path, 57)
      with self.turn_back_the_time(2*86400):
         rc= main.main(["test", "-f", self.configfile, "--backup"])
         self.assertEqual(rc, 0)
      runs= self.executesql("select run_name from run")
      self.assertEqual(len(runs), 1)
      run0= runs[0]

      with self.turn_back_the_time(86400):
         rc= main.main(["test", "-f", self.configfile, "--backup"])
         self.assertEqual(rc, 0)
      runs= self.executesql("select run_name from run")
      self.assertEqual(len(runs), 2)
      runs.remove(run0)
      run1= runs[0]

      rc= main.main(["test", "-f", self.configfile, "--backup"])
      self.assertEqual(rc, 0)
      runs= self.executesql("select run_name from run")
      self.assertEqual(len(runs), 2)
      runs.remove(run0)
      self.assertNotEqual(runs[0], run1)

   def test_purges_of_deleted_files(self):
      """
      when there is a backup in the oldest slot
      and the file is deleted
      and the deletion date maps to the oldest slot
      then there should not be any other backups
      and the oldest slot backup should be purged
      """
      self.setup_directories()
      file_a, file_b= self.homedir_abs("file_a", "file_b")
      with self.turn_back_the_time(160*DAY):
         t_a = self.create_file(file_a, 2)
         self.run_backup()
      with self.turn_back_the_time(159*DAY):
         os.unlink(file_a)
         t_b = self.create_file(file_b, 3)
         self.run_backup()
      self.run_backup()
      actual= set(self.executesql("select path, timestamp from content"))
      self.assertEqual(actual, {(file_b, t_b)})

   def test_non_deleted_files_dont_get_purged(self):
      """
      when there is a backup in the oldest slot
      but the file is not deleted
      then the oldest slot backup must not be purged
      """
      self.setup_directories()
      file_a, file_b= self.homedir_abs("file_a", "file_b")
      with self.turn_back_the_time(160*DAY):
         t_a = self.create_file(file_a, 2)
         self.run_backup()
      with self.turn_back_the_time(159*DAY):
         t_b = self.create_file(file_b, 3)
         self.run_backup()
      self.run_backup()
      actual= set(self.executesql("select path, timestamp from content"))
      self.assertEqual(actual, {(file_a,t_a), (file_b,t_b)})

   def test_deleted_file_only_last_backup_is_purged(self):
      """
      when there is a backup in the oldest slot
      and the file is deleted
      but the deletion date does not map to the oldest slot
      then the oldest slot backup must not be purged
      """
      self.setup_directories()
      file_a, file_b= self.homedir_abs("file_a", "file_b")
      with self.turn_back_the_time(160*DAY):
         t_a = self.create_file(file_a, 2)
         self.run_backup()
      with self.turn_back_the_time(50*DAY):
         os.unlink(file_a)
         t_b = self.create_file(file_b, 3)
         self.run_backup()
      self.run_backup()
      actual= set(self.executesql("select path, timestamp from content"))
      self.assertEqual(actual, {(file_a,t_a), (file_b,t_b)})

   def test_deletion_table_purge(self):
      """
      when the last backup of a path is purged
      then the deletion entry is also purged
      and checksum entry is also purged
      """
      self.setup_directories()
      file_a, file_b= self.homedir_abs("file_a", "file_b")
      with self.turn_back_the_time(160*DAY):
         self.create_file(file_a, 2)
         self.create_file(file_b, 3)
         self.run_backup()
      self.assertNotEqual(self.executesql("select * from last_checksummed where path=?",file_a), [])
      with self.turn_back_the_time(159*DAY):
         os.unlink(file_a)
         self.run_backup()
      with self.turn_back_the_time(50*DAY):
         os.unlink(file_b)
         self.run_backup()
      self.assertEqual(self.executesql("select path from deletion order by path"), [(file_a,), (file_b,)])
      self.run_backup()
      self.assertEqual(self.executesql("select path from deletion"), [(file_b,)])
      self.assertEqual(self.executesql("select * from last_checksummed where path=?",file_a), [])



# -*- coding: UTF-8 -*-
# Copyright © 2017 Cornelius Grotjahn - This file is covered by the LICENCE file in the project root directory.
import logging
import os
import re
from abus.testbase import AbusTestBase
from abus import backup
from abus import database
from abus import main

class UpgradeTests(AbusTestBase):
   def upgrade_and_check_tables_and_columns(self):
      # creation of database object causes the upgrade:
      with database.connect(self.databasepath, self.archivedir):
         pass
      with self.get_direct_db_connection() as conn:
         schema= {t[0]: set(c[1] for c in conn.execute("PRAGMA table_info('{}')".format(*t)))
                  for t in conn.execute("select name from sqlite_master where type='table'")}
         expected_schema= {
            "run": {"run_name", "archive_dir"},
            "content": {"run_name", "path", "timestamp", "checksum"},
            "location": {"checksum", "archive_dir", "is_compressed"},
            "last_checksummed": {"path", "st_dev", "st_ino", "mtime", "ctime"},
            "deletion": {"path", "timestamp"},
            "_database_version": {"current_version"},
         }
         self.assertDictEqual(schema, expected_schema)

         tbl= conn.execute("select * from _database_version").fetchall()
         self.assertEqual(tbl, [("4",)])
      del conn # otherwise file handle does not get released for some reason

   def test_create(self):
      self.setup_directories()
      self.upgrade_and_check_tables_and_columns()

   def setup_old_db(self, create_schema_version: int) -> None:
      """Creates an old, empty DB schema for given version"""
      self.setup_directories()
      os.unlink(self.databasepath)
      schema_str= None
      with open("{}/upgrade-{}.sql".format(os.path.dirname(__file__), create_schema_version)) as stream:
         for line in stream:
            if schema_str is None:
               if line.strip()=="EOF":
                  schema_str= ""
            else:
               schema_str += line
      with self.get_direct_db_connection() as conn:
         conn.executescript(schema_str)
         if create_schema_version>=3:
            conn.execute("""create table _database_version(current_version text)""")
            conn.execute("""insert into  _database_version values('{}.0')""".format(create_schema_version))

   def test_upgrade_schema_2_to_3(self):
      """Tests the v2->v3 transition where some new data has to be generated from the contents of the
      archive directory, namely is_compressed."""
      for from_version in (1,2):
         with self.subTest(from_version=from_version):
            self.setup_old_db(from_version)
            with self.get_direct_db_connection() as conn:
               if from_version==1:
                  conn.executemany("insert into run(run_name, timestamp)values(?,?)",
                                   [("2017_10_27_0810", 56), ("2015_06_08_1200", None)])
                  conn.executemany("insert into location(checksum,archive_dir) values(?,?)",
                                   [("32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2", "08/15"),
                                    ("f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa", "13/22")])
               else:
                  conn.execute("insert into completed_run(run_name) values(?)", ("2017_10_27_0810",))
                  conn.executemany("insert into content(run_name, path, timestamp, checksum) values(?,?,?,?)",
                                   [("2017_10_27_0810", "c:/some/path", 56, "32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2"),
                                    ("2015_06_08_1200", "C:/bla/bla",   97, "f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa"),
                                    ])
                  conn.executemany("insert into location(checksum,archive_dir) values(?,?)",
                                   [("32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2", "08/15"),
                                    ("f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa", "13/22")])
            del conn # otherwise file handle does not get released for some reason
            # generates a few archive files:
            for f in ["/47/11/2017_10_27_0810.lst",
                      "/34/15/2015_06_08_1200.lst",
                      "/08/15/32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2",
                      "/13/22/f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa.z",
                     ]:
               abspath= self.archivedir+f
               os.makedirs(os.path.dirname(abspath), exist_ok=True)
               with open(abspath, "w"): pass
            self.upgrade_and_check_tables_and_columns()
            tbl= self.executesql("select run_name, archive_dir from run")
            self.assertEqual(set(tbl), {("2017_10_27_0810","47/11"),("2015_06_08_1200","34/15")})
            tbl= self.executesql("select checksum, is_compressed from location")
            self.assertEqual(set(tbl), {("32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2", 0),
                                        ("f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa", 1),})

   def test_recover_from_upgrade_error(self):
      self.setup_old_db(2)
      with self.get_direct_db_connection() as conn:
         conn.execute("""drop table location""")
         conn.execute("""
            create table location(
               checksum text not null,
               archive_dir text not null)""")
         # introducing what after upgrade will be a unique constraint violation:
         conn.executemany("insert into location(checksum,archive_dir) values(?,?)",
                          [("32b772fb46ad237546d030bfd880d48220c57c555912f24873debfe8d0505ba2", "08/15"),
                           ("f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa", "13/22"),
                           ("f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa", "13/22"),
                           ])
      checksum_before= backup.calculate_checksum(self.databasepath)
      with self.assertLogs() as logs:
         rc= main.main(["test", "-f", self.configfile])
         self.assertEqual(rc, 4)
         self.assertTrue(any(r.levelno==logging.ERROR and "sqlite3.IntegrityError" in r.msg
                             for r in logs.records))
         self.assertTrue(any("index database upgrade failed, restoring backup" in r.msg
                             for r in logs.records))
      checksum_after= backup.calculate_checksum(self.databasepath)
      self.assertEqual(checksum_after, checksum_before)

   def test_upgrade_schema_3_to_4_checksum_cache_cleared(self):
      for from_version in range(1,4):
         with self.subTest(from_version=from_version):
            self.setup_old_db(from_version)
            with self.get_direct_db_connection() as conn:
               # old records must be deleted as part of the upgrade:
               conn.execute("insert into checksum_cache(dev, ino, timestamp, checksum) values(?,?,?,?)",
                            (23, 2**63-1, 1500000000, "f8e920545e99cdc9bbc2650eb8282344e8971a7ff0c397c91355d0fcaf6c61fa"))
            self.upgrade_and_check_tables_and_columns()
            del conn # otherwise file handle does not get released for some reason

   def test_upgrade_schema_3_to_4_superfluous_content_rows(self):
      data= [
         ("c:/home/file_a", "2017_11_01_0001", 1500000000, "32b772"),
         ("c:/home/file_a", "2017_11_01_0002", 1500000000, "a9b773"),
         ("c:/home/file_a", "2017_11_01_0003", 1600000000, "a9b773"),
         ("c:/home/file_a", "2017_11_01_0004", 1500000000, "32b772"),
         ("c:/home/file_a", "2017_11_01_0005", 1600000000, "8971a4"),
         ("c:/home/file_a", "2017_11_01_0006", 1600000000, "8971a4"),
         ("c:/home/file_a", "2017_11_01_0007", 1600000000, "8971a4"),
         ("c:/home/file_a", "2017_11_01_0008", 1650000000, "f24873"),

         ("c:/home/file_b", "2017_11_01_0001", 1400000001, "321b772"),
         ("c:/home/file_b", "2017_11_01_0002", 1500000002, "a92b773"),
         ("c:/home/file_b", "2017_11_01_0003", 1600000003, "a93b773"),
         ("c:/home/file_b", "2017_11_01_0004", 1500000004, "324b772"),
         ("c:/home/file_b", "2017_11_01_0005", 1600000005, "89571a4"),
         ("c:/home/file_b", "2017_11_01_0006", 1600000006, "89671a4"),
         ("c:/home/file_b", "2017_11_01_0007", 1600000007, "89771a4"),
         ("c:/home/file_b", "2017_11_01_0008", 1650000008, "f284873"),

         ("c:/home/file_c", "2017_11_01_0008", 1650200008, "1355d0f"),

         ("c:/home/file_d", "2017_11_01_0001", 1650200008, "1355d0f"),
         ("c:/home/file_d", "2017_11_01_0003", 1650200008, "1355d0f"),
         ("c:/home/file_d", "2017_11_01_0005", 1650200008, "1355d0f"),
         ("c:/home/file_d", "2017_11_01_0006", 1650200008, "1355d0f"),
         ]
      expected_data= set()
      prev= (None, None, None)
      for p,r,t,c in data:
         if (p,t,c) != prev:
            expected_data.add((p,r,t,c))
            prev= (p,t,c)
      for from_version in range(1,4):
         with self.subTest(from_version=from_version):
            self.setup_old_db(from_version)
            with self.get_direct_db_connection() as conn:
               conn.executemany("insert into content(path, run_name, timestamp, checksum) values(?,?,?,?)", data)
            del conn # otherwise file handle does not get released for some reason
            self.upgrade_and_check_tables_and_columns()
            actual_data= self.executesql("select path, run_name, timestamp, checksum from content")
            self.assertEqual(set(actual_data), expected_data)
